from dataclasses import dataclass, field
from typing import Dict

from aquiche.errors import InvalidExpressionError
from aquiche.utils._scanner import Scanner, TokenType


@dataclass
class SumExpressionParserConfig:
    case_sensitive: bool = False
    value_mapping: Dict[str, int] = field(default_factory=lambda: {})


class SumExpressionParser:
    config: SumExpressionParserConfig
    scanner: Scanner

    def __init__(self, config: SumExpressionParserConfig) -> None:
        self.config = config
        if not self.config.case_sensitive:
            self.config.value_mapping = {k.lower(): v for k, v in self.config.value_mapping.items()}
        self.scanner = Scanner()

    def parse(self, input_string: str) -> int:
        self.scanner.start_scan(input_string)
        result = 0
        while self.scanner.token_type != TokenType.TERMINATOR:
            num_value = self.__evaluate_number()
            identifier_value = self.__evaluate_identifier()
            result += num_value * identifier_value
        return result

    def __evaluate_identifier(self) -> int:
        if self.scanner.token_type != TokenType.IDENTIFIER:
            raise InvalidExpressionError(
                expression=self.scanner.input, position=self.scanner.position, error_message="expected identifier"
            )

        token = self.scanner.token if self.config.case_sensitive else self.scanner.token.lower()
        if not token in self.config.value_mapping:
            raise InvalidExpressionError(
                expression=self.scanner.input, position=self.scanner.position, error_message="unknown identifier"
            )
        value = self.config.value_mapping[token]
        self.scanner.scan()
        return value

    def __evaluate_number(self) -> int:
        if self.scanner.token_type != TokenType.NUMBER:
            raise InvalidExpressionError(
                expression=self.scanner.input, position=self.scanner.position, error_message="expected number"
            )

        value = int(self.scanner.token)
        self.scanner.scan()
        return value



# Stolen from pydantic who stole it from django

import re
from datetime import date, datetime, time, timedelta, timezone
from typing import Dict, Optional, Type, Union

from aquiche import errors
from aquiche.utils._sum_expression_parser import SumExpressionParser, SumExpressionParserConfig

DATE_EXPR = r"(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})"
TIME_EXPR = (
    r"(?P<hour>\d{1,2}):(?P<minute>\d{1,2})"
    r"(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?"
    r"(?P<tzinfo>Z|[+-]\d{2}(?::?\d{2})?)?$"
)

date_re = re.compile(f"{DATE_EXPR}$")
time_re = re.compile(TIME_EXPR)
datetime_re = re.compile(f"{DATE_EXPR}[T ]{TIME_EXPR}")

standard_duration_re = re.compile(
    r"^"
    r"(?:(?P<days>-?\d+) (days?, )?)?"
    r"((?:(?P<hours>-?\d+):)(?=\d+:\d+))?"
    r"(?:(?P<minutes>-?\d+):)?"
    r"(?P<seconds>-?\d+)"
    r"(?:\.(?P<microseconds>\d{1,6})\d{0,6})?"
    r"$"
)

# Support the sections of ISO 8601 date representation that are accepted by timedelta
iso8601_duration_re = re.compile(
    r"^(?P<sign>[-+]?)"
    r"P"
    r"(?:(?P<days>\d+(.\d+)?)D)?"
    r"(?:T"
    r"(?:(?P<hours>\d+(.\d+)?)H)?"
    r"(?:(?P<minutes>\d+(.\d+)?)M)?"
    r"(?:(?P<seconds>\d+(.\d+)?)S)?"
    r")?"
    r"$"
)

custom_duration_parser = SumExpressionParser(
    SumExpressionParserConfig(
        case_sensitive=False,
        value_mapping={
            "second": 1,
            "seconds": 1,
            "s": 1,
            "minute": 60,
            "minutes": 60,
            "m": 60,
            "hour": 3600,
            "hours": 3600,
            "h": 3600,
            "day": 24 * 3600,
            "days": 24 * 3600,
            "d": 24 * 3600,
        },
    )
)

EPOCH = datetime(1970, 1, 1)
# if greater than this, the number is in ms, if less than or equal it's in seconds
# (in seconds this is 11th October 2603, in ms it's 20th August 1970)
MS_WATERSHED = int(2e10)
# slightly more than datetime.max in ns - (datetime.max - EPOCH).total_seconds() * 1e9
MAX_NUMBER = int(3e20)
StrBytesIntFloat = Union[str, bytes, int, float]


def get_numeric(value: StrBytesIntFloat, native_expected_type: str) -> Union[None, int, float]:
    if isinstance(value, (int, float)):
        return value
    try:
        return float(value)
    except ValueError:
        return None
    except TypeError as err:
        raise TypeError(f"invalid type; expected {native_expected_type}, string, bytes, int or float") from err


def from_unix_seconds(seconds: Union[int, float]) -> datetime:
    if seconds > MAX_NUMBER:
        return datetime.max
    if seconds < -MAX_NUMBER:
        return datetime.min

    while abs(seconds) > MS_WATERSHED:
        seconds /= 1000
    parsed_date = EPOCH + timedelta(seconds=seconds)
    return parsed_date.replace(tzinfo=timezone.utc)


def _parse_timezone(value: Optional[str], error: Type[Exception]) -> Union[None, int, timezone]:
    if value == "Z":
        return timezone.utc
    if value is not None:
        offset_mins = int(value[-2:]) if len(value) > 3 else 0
        offset = 60 * int(value[1:3]) + offset_mins
        if value[0] == "-":
            offset = -offset
        try:
            return timezone(timedelta(minutes=offset))
        except ValueError as err:
            raise error(value) from err
    return None


def parse_date(value: Union[date, StrBytesIntFloat]) -> date:
    if isinstance(value, date):
        if isinstance(value, datetime):
            return value.date()
        return value

    number = get_numeric(value, "date")
    if number is not None:
        return from_unix_seconds(number).date()

    if isinstance(value, bytes):
        value = value.decode()

    match = date_re.match(value)  # type: ignore
    if match is None:
        raise errors.DateError(value)

    date_params = {k: int(v) for k, v in match.groupdict().items()}

    try:
        return date(**date_params)
    except ValueError as err:
        raise errors.DateError(value) from err


def parse_time(value: Union[time, StrBytesIntFloat]) -> time:
    if isinstance(value, time):
        return value

    number = get_numeric(value, "time")
    if number is not None:
        if number >= 86400:
            # doesn't make sense since the time time loop back around to 0
            raise errors.TimeError(value)
        return (datetime.min + timedelta(seconds=number)).time()

    if isinstance(value, bytes):
        value = value.decode()

    match = time_re.match(value)  # type: ignore
    if match is None:
        raise errors.TimeError(value)

    parsed_params = match.groupdict()
    if parsed_params["microsecond"]:
        parsed_params["microsecond"] = parsed_params["microsecond"].ljust(6, "0")

    tzinfo = _parse_timezone(parsed_params.pop("tzinfo"), errors.TimeError)
    time_params: Dict[str, Union[None, int, timezone]] = {k: int(v) for k, v in parsed_params.items() if v is not None}
    time_params["tzinfo"] = tzinfo

    try:
        return time(**time_params)  # type: ignore
    except ValueError as err:
        raise errors.TimeError(value) from err


def parse_datetime(value: Union[datetime, StrBytesIntFloat]) -> datetime:
    if isinstance(value, datetime):
        return value

    number = get_numeric(value, "datetime")
    if number is not None:
        return from_unix_seconds(number)

    if isinstance(value, bytes):
        value = value.decode()

    match = datetime_re.match(value)  # type: ignore
    if match is None:
        raise errors.DateTimeError(value)

    parsed_params = match.groupdict()
    if parsed_params["microsecond"]:
        parsed_params["microsecond"] = parsed_params["microsecond"].ljust(6, "0")

    tzinfo = _parse_timezone(parsed_params.pop("tzinfo"), errors.DateTimeError)
    datetime_params: Dict[str, Union[None, int, timezone]] = {
        k: int(v) for k, v in parsed_params.items() if v is not None
    }
    datetime_params["tzinfo"] = tzinfo

    try:
        return datetime(**datetime_params)  # type: ignore
    except ValueError as err:
        raise errors.DateTimeError(value) from err


def parse_duration(value: StrBytesIntFloat) -> timedelta:
    if isinstance(value, timedelta):
        return value

    if isinstance(value, (int, float)):
        # below code requires a string
        value = f"{value:f}"
    elif isinstance(value, bytes):
        value = value.decode()

    try:
        match = standard_duration_re.match(value) or iso8601_duration_re.match(value)
    except TypeError as err:
        raise TypeError("invalid type; expected timedelta, string, bytes, int or float") from err

    if not match:
        return __parse_duration_custom(value)

    parsed_params = match.groupdict()
    sign = -1 if parsed_params.pop("sign", "+") == "-" else 1
    if parsed_params.get("microseconds"):
        parsed_params["microseconds"] = parsed_params["microseconds"].ljust(6, "0")

    if parsed_params.get("seconds") and parsed_params.get("microseconds") and parsed_params["seconds"].startswith("-"):
        parsed_params["microseconds"] = "-" + parsed_params["microseconds"]

    timedelta_params = {k: float(v) for k, v in parsed_params.items() if v is not None}

    return sign * timedelta(**timedelta_params)


def __parse_duration_custom(value: str) -> timedelta:
    try:
        duration_seconds = custom_duration_parser.parse(value)
        return timedelta(seconds=duration_seconds)
    except Exception as err:
        raise errors.DurationError(value) from err



from curses.ascii import isalpha, isdigit, isspace
from enum import Enum


TERMINATOR = "\0"


class TokenType(str, Enum):
    NONE = "NONE"
    NUMBER = "NUMBER"
    IDENTIFIER = "IDENTIFIER"
    TERMINATOR = "TERMINATOR"


class Scanner:
    input: str
    look: str
    index: int
    position: int
    token: str
    token_type: TokenType

    def __init__(self) -> None:
        self.input = ""
        self.look = ""
        self.index = 0
        self.token = ""
        self.token_type = TokenType.NONE
        self.position = 0

    def start_scan(self, input_string: str) -> None:
        self.input = input_string.strip()
        self.index = 0
        self.__next()
        self.scan()

    def scan(self) -> None:
        while isspace(self.look):
            self.__next()

        self.token = ""
        self.position = self.index - 1

        if isalpha(self.look):
            while True:
                self.token += self.look
                self.__next()
                if not isalpha(self.look):
                    break
            self.token_type = TokenType.IDENTIFIER
            return

        if isdigit(self.look):
            while True:
                self.token += self.look
                self.__next()
                if not isdigit(self.look):
                    break
            self.token_type = TokenType.NUMBER
            return

        if self.look == TERMINATOR:
            self.token_type = TokenType.TERMINATOR
            return

        self.token_type = TokenType.NONE
        self.__next()

    def __next(self):
        if self.index >= len(self.input):
            self.look = TERMINATOR
        else:
            self.look = self.input[self.index]
            self.index += 1



from asyncio import Lock, Event, sleep as asleep
from datetime import datetime, timezone
import random
from time import sleep
from typing import Any, Awaitable, Callable, Tuple, Union

from aquiche import errors
from aquiche._core import CachedValue, CacheTaskExecutionInfo
from aquiche._expiration import (
    AsyncCacheExpiration,
    CacheExpiration,
)
from aquiche.utils._async_utils import AsyncWrapperMixin


class SyncCachedRecord:
    __get_function: Callable[..., Any]
    __get_exec_info: CacheTaskExecutionInfo
    __cached_value: CachedValue
    __expiration: CacheExpiration
    __negative_expiration: CacheExpiration

    def __init__(
        self,
        get_function: Callable[..., Any],
        get_exec_info: CacheTaskExecutionInfo,
        expiration: Union[CacheExpiration, AsyncCacheExpiration],
        negative_expiration: Union[CacheExpiration, AsyncCacheExpiration],
    ) -> None:
        expiration, negative_expiration = self.__validate_expirations(expiration, negative_expiration)
        self.__get_function = get_function  # type: ignore
        self.__get_exec_info = get_exec_info
        self.__cached_value = CachedValue()
        self.__expiration = expiration
        self.__negative_expiration = negative_expiration

    def get_cached(self) -> Any:
        if self.__cached_value.last_fetched is not None:
            if not self.is_expired():
                return self.__cached_value.value

        self.__store_cache()

        return self.__cached_value.value

    def destroy(self) -> None:
        self.__cached_value.destroy_value()

    def is_expired(self) -> bool:
        if self.__cached_value.last_fetched is None:
            return False
        if self.__cached_value.is_error:
            return self.__negative_expiration.is_value_expired(self.__cached_value)

        return self.__expiration.is_value_expired(self.__cached_value)

    def __validate_expirations(
        self,
        expiration: Union[CacheExpiration, AsyncCacheExpiration],
        negative_expiration: Union[CacheExpiration, AsyncCacheExpiration],
    ) -> Tuple[CacheExpiration, CacheExpiration]:
        error_messages = []
        if isinstance(expiration, AsyncCacheExpiration):
            error_messages.append("invalid expiration, use values that evaluate to 'sync' objects")
        if isinstance(negative_expiration, AsyncCacheExpiration):
            error_messages.append("invalid negative expiration, use values that evaluate to 'sync' objects")

        if len(error_messages) > 0:
            raise errors.InvalidCacheConfig(error_messages)
        return expiration, negative_expiration  # type: ignore

    def __store_cache(self) -> None:
        value, is_successful = self.__execute_task()

        self.__cached_value.last_fetched = datetime.now(timezone.utc)
        self.__cached_value.value = value
        self.__cached_value.is_error = not is_successful

        if not is_successful and self.__get_exec_info.fail:
            raise value

    def __execute_task(self) -> Tuple[Any, bool]:
        retry_iter = 0
        while True:
            try:
                return (self.__get_function(), True)
            except Exception as err:
                if retry_iter >= self.__get_exec_info.retries:
                    return err, False

                if self.__get_exec_info.backoff_in_seconds != 0:
                    sleep_seconds = self.__get_exec_info.backoff_in_seconds * 2**retry_iter + random.uniform(0, 1)
                    sleep(sleep_seconds)

                retry_iter += 1


class AsyncCachedRecord(AsyncWrapperMixin):
    __lock: Lock
    __get_function: Callable[..., Awaitable[Any]]
    __get_exec_info: CacheTaskExecutionInfo
    __cached_value: CachedValue
    __expiration: Union[CacheExpiration, AsyncCacheExpiration]
    __negative_expiration: Union[CacheExpiration, AsyncCacheExpiration]

    def __init__(
        self,
        get_function: Callable[..., Awaitable[Any]],
        get_exec_info: CacheTaskExecutionInfo,
        expiration: Union[AsyncCacheExpiration, CacheExpiration],
        negative_expiration: Union[AsyncCacheExpiration, CacheExpiration],
    ) -> None:
        self.__lock = Lock()
        self.__get_function = get_function  # type: ignore
        self.__get_exec_info = get_exec_info
        self.__cached_value = CachedValue()
        self.__expiration = expiration
        self.__negative_expiration = negative_expiration

    async def get_cached(self) -> Any:
        event = None
        await self.__lock.acquire()

        if self.__cached_value.last_fetched is not None:
            if not await self.is_expired():
                self.__lock.release()
                return self.__cached_value.value

        if self.__cached_value.inflight is not None:
            event = self.__cached_value.inflight
            self.__lock.release()
        else:
            self.__cached_value.inflight = Event()
            event = self.__cached_value.inflight
            self.__lock.release()

            await self.__store_cache()

        await event.wait()

        return self.__cached_value.value

    async def destroy(self) -> None:
        if self.__cached_value.last_fetched is None:
            return

        if self.__cached_value.exit_stack is not None:
            await self.__cached_value.exit_stack.aclose()
            self.__cached_value.exit_stack = None

        self.__cached_value.destroy_value()

    async def is_expired(self) -> bool:
        if self.__cached_value.last_fetched is None:
            return False

        expiration = None
        if self.__cached_value.is_error:
            expiration = self.__negative_expiration
        else:
            expiration = self.__expiration

        if isinstance(expiration, AsyncCacheExpiration):
            return await expiration.is_value_expired(self.__cached_value)
        return expiration.is_value_expired(self.__cached_value)

    async def __store_cache(self) -> None:
        if self.__cached_value.inflight is None:
            raise errors.DeadlockError()
        value, is_successful = await self.__execute_task()

        async with self.__lock:
            await self.destroy()
            event = self.__cached_value.inflight
            self.__cached_value.last_fetched = datetime.now(timezone.utc)
            self.__cached_value.inflight = None
            if is_successful:
                value, is_successful = await self.__safe_wrap_exit_stack(value)
            self.__cached_value.value = value
            self.__cached_value.is_error = not is_successful
            event.set()

        if not is_successful and self.__get_exec_info.fail:
            raise value

    async def __execute_task(self) -> Tuple[Any, bool]:
        retry_iter = 0
        while True:
            try:
                return (await self.__get_function(), True)
            except Exception as err:
                if retry_iter >= self.__get_exec_info.retries:
                    return err, False

                if self.__get_exec_info.backoff_in_seconds != 0:
                    sleep_seconds = self.__get_exec_info.backoff_in_seconds * 2**retry_iter + random.uniform(0, 1)
                    await asleep(sleep_seconds)

                retry_iter += 1

    async def __safe_wrap_exit_stack(self, value: Any) -> Tuple[Any, bool]:
        try:
            exit_stack, value = await self.wrap_async_exit_stack(
                value=value, wrap_config=self.__get_exec_info.wrap_async_exit_stack
            )
            self.__cached_value.exit_stack = exit_stack
            return value, True
        except Exception as err:
            return err, False



from dataclasses import dataclass
import functools
from typing import Any, Dict

from aquiche import errors


@dataclass
class MissingValue:
    default_value: Any = None


def __rgetattr(obj: Any, path: str, default: Any = None) -> Any:
    attrs = path.split(".")
    try:
        return functools.reduce(getattr, attrs, obj)
    except AttributeError:
        return default


def __rsetattr(obj: Any, path: str, value: Any) -> None:
    current_attr, _, leftover_path = path.rpartition(".")
    return setattr(__rgetattr(obj=obj, path=current_attr) if current_attr else obj, leftover_path, value)


def __deep_get(dictionary: Dict, key_path: str, default: Any = None) -> Any:
    return functools.reduce(
        lambda d, key: d.get(key, default) if isinstance(d, dict) else MissingValue(default_value=default),
        key_path.split("."),
        dictionary,
    )


def __deep_set(dictionary: Dict, key_path: str, value: Any) -> None:
    dict_iter = dictionary
    key_iter = key_path.split(".")
    for key in key_iter[:-1]:
        dict_iter = dict_iter.setdefault(key, {})
    dict_iter[key_iter[-1]] = value


def extract_from_obj(
    obj: Any, attribute_path: str, check_attribute_exists: bool = True, default_value: Any = None
) -> Any:
    if not isinstance(attribute_path, str):
        raise errors.ExtractionError(attribute_path)
    attribute_path = attribute_path.strip().lstrip("$.")
    value = MissingValue(default_value)
    if isinstance(obj, dict):
        value = __deep_get(dictionary=obj, key_path=attribute_path, default=MissingValue(default_value))
    else:
        value = __rgetattr(obj=obj, path=attribute_path, default=MissingValue(default_value))
    if isinstance(value, MissingValue):
        if check_attribute_exists:
            raise errors.ExtractionError(attribute_path)
        return value.default_value

    return value


def set_value_obj(obj: Any, attribute_path: str, value: Any) -> None:
    if not isinstance(attribute_path, str):
        raise errors.ExtractionError(attribute_path)
    attribute_path = attribute_path.strip().lstrip("$.")
    if isinstance(obj, dict):
        return __deep_set(dictionary=obj, key_path=attribute_path, value=value)
    return __rsetattr(obj=obj, path=attribute_path, value=value)



from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Union, get_args

from aquiche.errors import InvalidCacheConfig
from aquiche._expiration import CacheExpirationValue, DurationExpirationValue


@dataclass
class CacheParameters:
    enabled: bool = False
    maxsize: Optional[int] = None
    expiration: Optional[CacheExpirationValue] = None
    expired_items_auto_removal_period: Optional[DurationExpirationValue] = None
    wrap_async_exit_stack: Union[bool, List[str], None] = None
    negative_cache: bool = False
    negative_expiration: Optional[CacheExpirationValue] = None
    retry_count: int = 0
    backoff_in_seconds: Union[int, float] = 0


def __extract_type_names(types: Tuple[Any, ...]) -> str:
    return "|".join(map(lambda t: t.__name__, types))


def validate_cache_params(
    enabled: bool,
    maxsize: Optional[int],
    expiration: Optional[CacheExpirationValue],
    expired_items_auto_removal_period: Optional[DurationExpirationValue],
    wrap_async_exit_stack: Union[bool, List[str], None],
    negative_cache: bool,
    negative_expiration: Optional[CacheExpirationValue],
    retry_count: int,
    backoff_in_seconds: Union[int, float],
) -> None:
    errors = []
    if not isinstance(enabled, bool):
        errors += ["enabled should be either bool or a callable function"]
    if not (maxsize is None or isinstance(maxsize, int)):
        errors += ["maxsize should be int or None"]
    if not (expiration is None or isinstance(expiration, get_args(CacheExpirationValue))):
        errors += [
            f"expiration should be either None or one of these types: {__extract_type_names(get_args(CacheExpirationValue))}"
        ]
    if not (negative_expiration is None or isinstance(negative_expiration, get_args(CacheExpirationValue))):
        errors += [
            f"negative expiration should be either None or one of these types: {__extract_type_names(get_args(CacheExpirationValue))}"
        ]
    if not (
        expired_items_auto_removal_period is None
        or isinstance(expired_items_auto_removal_period, get_args(DurationExpirationValue))
    ):
        errors += [
            "expired_items_auto_removal_period should be either None or one of these types:"
            + __extract_type_names(get_args(DurationExpirationValue))
        ]
    if not (
        wrap_async_exit_stack is None
        or isinstance(wrap_async_exit_stack, bool)
        or (
            isinstance(wrap_async_exit_stack, list)
            and all((isinstance(wrapper, str) for wrapper in wrap_async_exit_stack))
        )
    ):
        errors += ["wrap_async_exit_stack should be either None, bool or a callable function"]

    if not isinstance(negative_cache, bool):
        errors += ["negative_cache should be bool"]
    if not isinstance(retry_count, int):
        errors += ["retry_count should be an integer"]
    if not isinstance(backoff_in_seconds, (int, float)):
        errors += ["backoff_in_seconds should be a number"]

    if errors:
        raise InvalidCacheConfig(errors)



from typing import Any, Callable, Tuple


class HashedSeq(list):
    hash_value: str

    def __init__(self, tup: Tuple, hash_fn: Callable = hash):
        self[:] = tup
        self.hash_value = hash_fn(tup)

    def __hash__(self):
        return self.hash_value


def __make_key_from_args(args: Any, kwargs: Any, kwd_mark=(object(),)) -> Any:
    # All of code below relies on kwargs preserving the order input by the user.
    # Formerly, we sorted() the kwargs before looping.  The new way is *much*
    # faster; however, it means that f(x=1, y=2) will now be treated as a
    # distinct call from f(y=2, x=1) which will be cached separately.
    fast_types = ({int, str},)

    key = args
    if kwargs:
        key += kwd_mark
        for item in kwargs.items():
            key += item

    key += tuple(type(v) for v in args)
    if kwargs:
        key += tuple(type(v) for v in kwargs.values())
    elif len(key) == 1 and type(key[0]) in fast_types:
        return key[0]
    return HashedSeq(key)


def make_key(*args: Any, **kwargs: Any) -> Any:
    return __make_key_from_args(args, kwargs)



from asyncio import (
    gather,
    iscoroutinefunction,
    Lock,
)
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta, timezone
from functools import partial, update_wrapper
import sys
from threading import RLock
from typing import Awaitable, Callable, Iterable, List, Optional, Protocol, TypeVar, Union

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

from aquiche._cache import AsyncCachedRecord, SyncCachedRecord
from aquiche._cache_params import CacheParameters, validate_cache_params
from aquiche._core import CacheTaskExecutionInfo
from aquiche.errors import InvalidCacheConfig
from aquiche._expiration import (
    CacheExpirationValue,
    DurationExpirationValue,
    get_cache_expiration,
    NonExpiringCacheExpiration,
)
from aquiche._hash import make_key
from aquiche._repository import CacheRepository, LRUCacheRepository
from aquiche.utils._async_utils import awaitify
from aquiche.utils._time_parse import parse_duration
from aquiche.utils._singleton import Singleton

T = TypeVar("T")
P = ParamSpec("P")
C = TypeVar("C", bound=Callable)


@dataclass
class CacheInfo:
    hits: int = 0
    misses: int = 0
    maxsize: Optional[int] = None
    current_size: int = 0
    last_expiration_check: Optional[datetime] = None


class AquicheFunctionWrapper(Protocol[C]):
    cache_info: Union[Callable[..., CacheInfo], Callable[..., Awaitable[CacheInfo]]]
    clear_cache: Union[Callable[..., None], Callable[..., Awaitable[None]]]
    cache_parameters: Callable[..., CacheParameters]
    remove_expired: Union[Callable[..., None], Callable[..., Awaitable[None]]]

    __call__: C


def __parse_duration_to_timedelta(duration: Optional[DurationExpirationValue]) -> Optional[timedelta]:
    if duration is None:
        return None
    if isinstance(duration, timedelta):
        return duration
    return parse_duration(duration)


class CacheCleanupRegistry(metaclass=Singleton):
    __clear_callbacks: List[Union[Callable[..., None], Callable[..., Awaitable[None]]]]

    def __init__(self) -> None:
        self.__clear_callbacks = []

    def register_clear_callback(
        self, clear_callback: Union[Callable[..., None], Callable[..., Awaitable[None]]]
    ) -> None:
        self.__clear_callbacks.append(clear_callback)

    def get_clear_callbacks(self) -> Iterable[Union[Callable[..., None], Callable[..., Awaitable[None]]]]:
        return iter(self.__clear_callbacks)

    def get_async_callbacks(
        self, callbacks: Iterable[Union[Callable[..., None], Callable[..., Awaitable[None]]]]
    ) -> Iterable[Callable[..., Awaitable[None]]]:
        return (
            callback_function if iscoroutinefunction(callback_function) else awaitify(callback_function)  # type: ignore
            for callback_function in callbacks
        )


def alru_cache(
    __func: Union[Callable[P, T], None] = None,
    enabled: bool = True,
    maxsize: Optional[int] = None,
    expiration: Optional[CacheExpirationValue] = None,
    expired_items_auto_removal_period: Optional[DurationExpirationValue] = None,
    wrap_async_exit_stack: Union[bool, List[str], None] = None,
    negative_cache: bool = False,
    negative_expiration: Optional[CacheExpirationValue] = "10 seconds",
    retry_count: int = 0,
    backoff_in_seconds: Union[int, float] = 0,
) -> AquicheFunctionWrapper[Callable[P, T]]:
    validate_cache_params(
        enabled=enabled,
        maxsize=maxsize,
        expiration=expiration,
        expired_items_auto_removal_period=expired_items_auto_removal_period,
        wrap_async_exit_stack=wrap_async_exit_stack,
        negative_cache=negative_cache,
        negative_expiration=negative_expiration,
        retry_count=retry_count,
        backoff_in_seconds=backoff_in_seconds,
    )
    cache_params = CacheParameters(
        enabled=enabled,
        maxsize=maxsize,
        expiration=expiration,
        expired_items_auto_removal_period=expired_items_auto_removal_period,
        wrap_async_exit_stack=wrap_async_exit_stack,
        negative_cache=negative_cache,
        negative_expiration=negative_expiration,
        retry_count=retry_count,
        backoff_in_seconds=backoff_in_seconds,
    )
    if maxsize is not None:
        # Negative maxsize is treated as 0
        maxsize = max(maxsize, 0)
    # Negative retry count is treated as 0
    retry_count = max(retry_count, 0)

    if __func is not None and callable(__func):
        # The user_function was passed in directly via the hidden __func argument
        user_function = __func
        if iscoroutinefunction(user_function):
            wrapper = _async_lru_cache_wrapper(
                user_function=user_function,
                **asdict(cache_params),
            )
        else:
            wrapper = _sync_lru_cache_wrapper(
                user_function=user_function,
                **asdict(cache_params),
            )
        wrapper.cache_parameters = lambda: cache_params  # type: ignore
        return update_wrapper(wrapper, user_function)  # type: ignore

    def decorating_function(user_function: Union[Callable[P, T], Callable[P, Awaitable[T]]]):
        if iscoroutinefunction(user_function):
            wrapper = _async_lru_cache_wrapper(
                user_function=user_function,
                **asdict(cache_params),
            )
        else:
            wrapper = _sync_lru_cache_wrapper(
                user_function=user_function,
                **asdict(cache_params),
            )
        wrapper.cache_parameters = lambda: cache_params  # type: ignore
        return update_wrapper(wrapper, user_function)

    return decorating_function  # type: ignore


async def clear_all() -> None:
    cleanup_repository = CacheCleanupRegistry()

    callback_functions = cleanup_repository.get_async_callbacks(cleanup_repository.get_clear_callbacks())
    await gather(*(callback_function() for callback_function in callback_functions))


def clear_all_sync() -> None:
    cleanup_repository = CacheCleanupRegistry()

    for clear_callback in cleanup_repository.get_clear_callbacks():
        if not iscoroutinefunction(clear_callback):
            clear_callback()


def _sync_lru_cache_wrapper(
    user_function: Callable[P, T],
    enabled: bool,
    maxsize: Optional[int],
    expiration: Optional[CacheExpirationValue],
    expired_items_auto_removal_period: Optional[DurationExpirationValue],
    wrap_async_exit_stack: Union[bool, List[str], None],
    negative_cache: bool,
    negative_expiration: Optional[CacheExpirationValue],
    retry_count: int,
    backoff_in_seconds: Union[int, float],
) -> AquicheFunctionWrapper[Callable[P, T]]:
    if wrap_async_exit_stack:
        raise InvalidCacheConfig(["wrap_async_exit_stack can only be used with async functions"])

    cache: CacheRepository = LRUCacheRepository(maxsize=maxsize)
    cleanup_repository = CacheCleanupRegistry()

    hits = misses = 0
    lock = RLock()  # because cache updates aren't thread-safe
    last_expiration_check = datetime.fromtimestamp(0, tz=timezone.utc)
    expiry_period = __parse_duration_to_timedelta(expired_items_auto_removal_period)

    def __is_cache_enabled() -> bool:
        if maxsize == 0:
            return False
        if callable(enabled):
            return enabled()
        return enabled

    def __remove_expired() -> None:
        nonlocal last_expiration_check
        last_expiration_check = datetime.now(timezone.utc)
        removed_items = cache.filter(lambda _key, record: not record.is_expired())
        for removed_item in removed_items:
            removed_item.destroy()

    def __schedule_remove_expired() -> None:
        if expiry_period is None:
            return

        if datetime.now(timezone.utc) - last_expiration_check >= expiry_period:
            __remove_expired()

    if not __is_cache_enabled():

        def wrapper(*args, **kwargs) -> T:
            # No caching -- just a statistics update
            nonlocal cache, misses
            misses += 1
            result = user_function(*args, **kwargs)
            return result

    elif maxsize is None:

        def wrapper(*args, **kwargs) -> T:
            # Simple caching without ordering or size limit
            nonlocal hits, misses
            key = make_key(*args, **kwargs)

            with lock:
                __schedule_remove_expired()

                record = cache.get_no_adjust(key)
                if record is not None:
                    hits += 1
                    return record.get_cached()
                misses += 1

            record = SyncCachedRecord(
                get_function=partial(user_function, *args, **kwargs),
                get_exec_info=CacheTaskExecutionInfo(
                    fail=not negative_cache,
                    retries=retry_count,
                    backoff_in_seconds=backoff_in_seconds,
                    wrap_async_exit_stack=False,
                ),
                expiration=get_cache_expiration(
                    expiration, prefer_async=False, default_expiration=NonExpiringCacheExpiration()
                ),
                negative_expiration=get_cache_expiration(
                    negative_expiration, prefer_async=False, default_expiration=NonExpiringCacheExpiration()
                ),
            )
            result = record.get_cached()

            with lock:
                cache.add_no_adjust(key=key, value=record)

            return result

    else:

        def wrapper(*args, **kwargs) -> T:
            # Size limited caching that tracks accesses by recency
            nonlocal hits, misses
            key = make_key(*args, **kwargs)

            with lock:
                __schedule_remove_expired()

                record = cache.get(key)
                if record is not None:
                    hits += 1
                    return record.get_cached()
                misses += 1

            record = SyncCachedRecord(
                get_function=partial(user_function, *args, **kwargs),
                get_exec_info=CacheTaskExecutionInfo(
                    fail=not negative_cache,
                    retries=retry_count,
                    backoff_in_seconds=backoff_in_seconds,
                    wrap_async_exit_stack=False,
                ),
                expiration=get_cache_expiration(
                    expiration, prefer_async=False, default_expiration=NonExpiringCacheExpiration()
                ),
                negative_expiration=get_cache_expiration(
                    negative_expiration, prefer_async=False, default_expiration=NonExpiringCacheExpiration()
                ),
            )
            result = record.get_cached()

            with lock:
                cache.add(key, record)

            return result

    def cache_info() -> CacheInfo:
        """Report cache statistics"""
        with lock:
            return CacheInfo(
                hits=hits,
                misses=misses,
                maxsize=maxsize,
                current_size=cache.get_size(),
                last_expiration_check=last_expiration_check,
            )

    def clear_cache() -> None:
        """Clear the cache and cache statistics"""
        nonlocal cache, hits, misses
        with lock:
            cache.every(lambda _key, value: value.destroy())
            cache.clear()
            hits = misses = 0

    def remove_expired() -> None:
        """Remove expired items from the cache"""
        with lock:
            __remove_expired()

    cleanup_repository.register_clear_callback(clear_cache)

    wrapper.cache_info = cache_info  # type: ignore
    wrapper.clear_cache = clear_cache  # type: ignore
    wrapper.cache_parameters = CacheParameters  # type: ignore
    wrapper.remove_expired = remove_expired  # type: ignore
    return wrapper  # type: ignore


def _async_lru_cache_wrapper(
    user_function: Callable[P, T],
    enabled: Union[bool, Callable[[], bool]],
    maxsize: Optional[int],
    expiration: Optional[CacheExpirationValue],
    expired_items_auto_removal_period: Union[str, bytes, int, float, timedelta, None],
    wrap_async_exit_stack: Union[bool, List[str], None],
    negative_cache: bool,
    negative_expiration: Optional[CacheExpirationValue],
    retry_count: int,
    backoff_in_seconds: Union[int, float],
) -> AquicheFunctionWrapper[Callable[P, T]]:
    cache: CacheRepository = LRUCacheRepository(maxsize=maxsize)
    cleanup_repository = CacheCleanupRegistry()

    hits = misses = 0
    lock = Lock()  # because cache updates aren't concurrency-safe
    last_expiration_check = datetime.fromtimestamp(0, tz=timezone.utc)
    expiry_period = __parse_duration_to_timedelta(expired_items_auto_removal_period)

    def __is_cache_enabled() -> bool:
        if maxsize == 0:
            return False
        if callable(enabled):
            return enabled()
        return enabled

    async def __expiry_filter_lambda(_key: str, record: AsyncCachedRecord) -> bool:
        return not await record.is_expired()

    async def __apply_destroy_lambda(_key: str, record: AsyncCachedRecord) -> None:
        await record.destroy()

    async def __remove_expired() -> None:
        nonlocal last_expiration_check
        last_expiration_check = datetime.now(timezone.utc)

        removed_items: List[AsyncCachedRecord] = await cache.filter_async(__expiry_filter_lambda)
        await gather(*(record.destroy() for record in removed_items))

    async def __schedule_remove_expired() -> None:
        nonlocal last_expiration_check

        if expiry_period is None:
            return

        if datetime.now(timezone.utc) - last_expiration_check >= expiry_period:
            await __remove_expired()

    if not __is_cache_enabled():

        async def wrapper(*args, **kwargs) -> T:
            # No caching -- just a statistics update
            nonlocal cache, misses
            misses += 1
            result = await user_function(*args, **kwargs)  # type: ignore
            return result

    elif maxsize is None:

        async def wrapper(*args, **kwargs) -> T:
            # Simple caching without ordering or size limit
            nonlocal hits, misses

            key = make_key(*args, **kwargs)

            async with lock:
                await __schedule_remove_expired()

                record = cache.get_no_adjust(key)
                if record is not None:
                    hits += 1
                    return await record.get_cached()
                misses += 1

            record = AsyncCachedRecord(
                get_function=partial(user_function, *args, **kwargs),  # type: ignore
                get_exec_info=CacheTaskExecutionInfo(
                    fail=not negative_cache,
                    retries=retry_count,
                    backoff_in_seconds=backoff_in_seconds,
                    wrap_async_exit_stack=wrap_async_exit_stack or False,
                ),
                expiration=get_cache_expiration(
                    expiration, prefer_async=True, default_expiration=NonExpiringCacheExpiration()
                ),
                negative_expiration=get_cache_expiration(
                    negative_expiration, prefer_async=True, default_expiration=NonExpiringCacheExpiration()
                ),
            )
            result = await record.get_cached()

            async with lock:
                cache.add_no_adjust(key=key, value=record)

            return result

    else:

        async def wrapper(*args, **kwargs) -> T:
            # Size limited caching that tracks accesses by recency
            nonlocal hits, misses
            key = make_key(*args, **kwargs)

            async with lock:
                await __schedule_remove_expired()

                result = cache.get(key)
                if result is not None:
                    hits += 1
                    return await result.get_cached()
                misses += 1

            record = AsyncCachedRecord(
                get_function=partial(user_function, *args, **kwargs),  # type: ignore
                get_exec_info=CacheTaskExecutionInfo(
                    fail=not negative_cache,
                    retries=retry_count,
                    backoff_in_seconds=backoff_in_seconds,
                    wrap_async_exit_stack=wrap_async_exit_stack or False,
                ),
                expiration=get_cache_expiration(
                    expiration, prefer_async=True, default_expiration=NonExpiringCacheExpiration()
                ),
                negative_expiration=get_cache_expiration(
                    negative_expiration, prefer_async=True, default_expiration=NonExpiringCacheExpiration()
                ),
            )
            result = await record.get_cached()

            async with lock:
                cache.add(key=key, value=record)

            return result

    async def cache_info() -> CacheInfo:
        """Report cache statistics"""
        async with lock:
            return CacheInfo(
                hits=hits,
                misses=misses,
                maxsize=maxsize,
                current_size=cache.get_size(),
                last_expiration_check=last_expiration_check,
            )

    async def clear_cache() -> None:
        """Clear the cache and cache statistics"""
        nonlocal cache, hits, misses
        async with lock:
            await cache.every_async(__apply_destroy_lambda)
            cache.clear()
            hits = misses = 0

    async def remove_expired() -> None:
        """Remove expired items from the cache"""
        async with lock:
            await __remove_expired()

    cleanup_repository.register_clear_callback(clear_cache)

    wrapper.cache_info = cache_info  # type: ignore
    wrapper.clear_cache = clear_cache  # type: ignore
    wrapper.cache_parameters = CacheParameters  # type: ignore
    wrapper.remove_expired = remove_expired  # type: ignore
    return wrapper  # type: ignore



from abc import ABCMeta, abstractmethod
from asyncio import iscoroutinefunction
from datetime import date, datetime, time, timedelta, timezone
from typing import Any, Awaitable, Callable, Coroutine, Optional, Union

from aquiche import errors
from aquiche._core import CachedValue, CachedItem
from aquiche.utils._async_utils import awaitify
from aquiche.utils._extraction_utils import extract_from_obj
from aquiche.utils._time_parse import parse_datetime, parse_date, parse_duration, parse_time

CacheExpirationValue = Union[bool, int, float, str, bytes, date, datetime, time, timedelta, Coroutine, Callable]
DurationExpirationValue = Union[str, bytes, int, float, timedelta]


def _get_cache_func_value(cached_value: CachedValue) -> CachedItem:
    assert cached_value.last_fetched is not None
    return CachedItem(value=cached_value.value, last_fetched=cached_value.last_fetched, is_error=cached_value.is_error)


class CacheExpiration(metaclass=ABCMeta):
    @abstractmethod
    def is_value_expired(self, value: CachedValue) -> bool:
        ...


class AsyncCacheExpiration(metaclass=ABCMeta):
    @abstractmethod
    async def is_value_expired(self, value: CachedValue) -> bool:
        ...


def _validate_sync_expiration(cache_expiration: Union[CacheExpiration, AsyncCacheExpiration], value: Any) -> None:
    if isinstance(cache_expiration, AsyncCacheExpiration):
        if (iscoroutinefunction(value) or callable(value)) and hasattr(value, "__name__"):
            value = str(value.__name__)

        raise errors.InvalidSyncExpirationType(value)


class NonExpiringCacheExpiration(CacheExpiration):
    def is_value_expired(self, value: CachedValue) -> bool:
        return value.last_fetched is None

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, NonExpiringCacheExpiration)


class BoolCacheExpiration(CacheExpiration):
    __is_expired: bool

    def __init__(self, is_expired: bool) -> None:
        super().__init__()
        self.__is_expired = is_expired

    @property
    def is_expired(self) -> bool:
        return self.__is_expired

    def is_value_expired(self, _value: CachedValue) -> bool:
        return self.is_expired

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, BoolCacheExpiration) and other.is_expired == self.is_expired


class DateCacheExpiration(CacheExpiration):
    __expiry_date: datetime

    def __init__(self, expiry_date: datetime) -> None:
        super().__init__()
        if expiry_date.tzinfo is not None and expiry_date.tzinfo.utcoffset(expiry_date) is not None:
            self.__expiry_date = expiry_date
        else:
            self.__expiry_date = expiry_date.replace(tzinfo=timezone.utc)

    @property
    def expiry_date(self) -> datetime:
        return self.__expiry_date

    def is_value_expired(self, value: CachedValue) -> bool:
        if value.last_fetched is None:
            return True
        return value.last_fetched >= self.expiry_date

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, DateCacheExpiration) and self.expiry_date == other.expiry_date


class RefreshingCacheExpiration(CacheExpiration):
    __refresh_interval: timedelta

    def __init__(self, refresh_interval: timedelta) -> None:
        super().__init__()
        self.__refresh_interval = refresh_interval

    @property
    def refresh_interval(self) -> timedelta:
        return self.__refresh_interval

    def is_value_expired(self, value: CachedValue) -> bool:
        if value.last_fetched is None:
            return True
        return (datetime.now(timezone.utc) - value.last_fetched) >= self.refresh_interval

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, RefreshingCacheExpiration) and self.refresh_interval == other.refresh_interval


class SyncAttributeCacheExpiration(CacheExpiration):
    __attribute_path: str

    def __init__(self, attribute_path: str) -> None:
        super().__init__()
        self.__attribute_path = attribute_path.strip()

    @property
    def attribute_path(self) -> str:
        return self.__attribute_path

    def is_value_expired(self, value: CachedValue) -> bool:
        expiry_value = extract_from_obj(obj=value.value, attribute_path=self.attribute_path)
        cache_expiration = get_cache_expiration(value=expiry_value, prefer_async=False)
        _validate_sync_expiration(cache_expiration=cache_expiration, value=expiry_value)
        return cache_expiration.is_value_expired(value)  # type: ignore

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, SyncAttributeCacheExpiration) and self.attribute_path == other.attribute_path


class SyncFuncCacheExpiration(CacheExpiration):
    __func: Callable[..., Any]

    def __init__(self, func: Callable[..., Any]) -> None:
        super().__init__()
        self.__func = func  # type: ignore

    @property
    def func(self) -> Callable[..., Any]:
        return self.__func

    def is_value_expired(self, value: CachedValue) -> bool:
        expiry_value = self.func(_get_cache_func_value(value))
        cache_expiration = get_cache_expiration(expiry_value, prefer_async=False)
        _validate_sync_expiration(cache_expiration=cache_expiration, value=expiry_value)
        return cache_expiration.is_value_expired(value)  # type: ignore

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, SyncFuncCacheExpiration) and self.func == other.func


class AsyncAttributeCacheExpiration(AsyncCacheExpiration):
    __attribute_path: str

    def __init__(self, attribute_path: str) -> None:
        super().__init__()
        self.__attribute_path = attribute_path.strip()

    @property
    def attribute_path(self) -> str:
        return self.__attribute_path

    async def is_value_expired(self, value: CachedValue) -> bool:
        expiry_value = extract_from_obj(obj=value.value, attribute_path=self.attribute_path)
        cache_expiration = get_cache_expiration(expiry_value)
        if isinstance(cache_expiration, CacheExpiration):
            return cache_expiration.is_value_expired(value)
        return await cache_expiration.is_value_expired(value)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, AsyncAttributeCacheExpiration) and self.attribute_path == other.attribute_path


class AsyncFuncCacheExpiration(AsyncCacheExpiration):
    __func: Callable[..., Awaitable[Any]]

    def __init__(self, func: Union[Callable[..., Any], Callable[..., Awaitable[Any]], Coroutine]) -> None:
        super().__init__()
        self.__func = awaitify(func)  # type: ignore

    @property
    def func(self) -> Callable[..., Awaitable[Any]]:
        return self.__func

    async def is_value_expired(self, value: CachedValue) -> bool:
        expiry_value = await self.__func(_get_cache_func_value(value))
        cache_expiration = get_cache_expiration(expiry_value)
        if isinstance(cache_expiration, CacheExpiration):
            return cache_expiration.is_value_expired(value)
        return await cache_expiration.is_value_expired(value)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, AsyncFuncCacheExpiration) and self.func == other.func


def get_cache_expiration(
    value: Optional[CacheExpirationValue],
    prefer_async: bool = True,
    default_expiration: Union[CacheExpiration, AsyncCacheExpiration, None] = None,
) -> Union[CacheExpiration, AsyncCacheExpiration]:
    if value is None:
        return default_expiration or NonExpiringCacheExpiration()
    if isinstance(value, bool):
        return BoolCacheExpiration(value)
    if isinstance(value, (float, int)):
        return __get_cache_expiration_from_num(value)
    if isinstance(value, (str, bytes)):
        return __get_cache_expiration_from_str(value=value.strip(), prefer_async=prefer_async)
    if isinstance(value, (date, datetime, time, timedelta)):
        return __get_cache_expiration_from_time(value)
    if iscoroutinefunction(value):
        return AsyncFuncCacheExpiration(value)
    if callable(value):
        return AsyncFuncCacheExpiration(value) if prefer_async else SyncFuncCacheExpiration(value)
    raise errors.InvalidExpirationType()


def __get_cache_expiration_from_num(value: Union[int, float]) -> CacheExpiration:
    # if the number is large enough we assume it's a timestamp
    value = int(value)
    if value > int(1e8):
        return DateCacheExpiration(expiry_date=parse_datetime(value))
    # otherwise assume it's a refresh interval in seconds
    return RefreshingCacheExpiration(refresh_interval=parse_duration(value))


def __get_cache_expiration_from_str(
    value: Union[str, bytes], prefer_async: bool
) -> Union[CacheExpiration, AsyncCacheExpiration]:
    value = __parse_value_to_str(value)

    if value.startswith("$."):
        return (
            AsyncAttributeCacheExpiration(attribute_path=value)
            if prefer_async
            else SyncAttributeCacheExpiration(attribute_path=value)
        )

    parsed_value: Any = None
    parse_functions = (parse_duration, parse_datetime, parse_date, parse_time)
    for parse_function in parse_functions:
        try:
            parsed_value = parse_function(value)
            break
        except Exception:
            pass

    if parsed_value is None:
        raise errors.InvalidTimeFormatError(value)

    return __get_cache_expiration_from_time(parsed_value)


def __get_cache_expiration_from_time(
    value: Union[date, datetime, time, timedelta]
) -> Union[CacheExpiration, AsyncCacheExpiration]:
    if isinstance(value, datetime):
        return DateCacheExpiration(expiry_date=value)

    if isinstance(value, date):
        return DateCacheExpiration(expiry_date=datetime.combine(value, datetime.min.time(), tzinfo=timezone.utc))

    if isinstance(value, time):
        return DateCacheExpiration(expiry_date=datetime.combine(date.today(), value))

    if isinstance(value, timedelta):
        return RefreshingCacheExpiration(refresh_interval=value)

    raise errors.InvalidTimeFormatError(value)


def __parse_value_to_str(value: Union[str, bytes]) -> str:
    if isinstance(value, str):
        return value
    return value.decode()



from asyncio import iscoroutinefunction
from contextlib import AsyncExitStack
from inspect import isawaitable
from typing import Any, Awaitable, Callable, Coroutine, List, Optional, Tuple, Union

from aquiche.utils._extraction_utils import extract_from_obj, set_value_obj


def awaitify(
    func: Union[Callable[..., Any], Callable[..., Awaitable[Any]], Coroutine]
) -> Callable[..., Awaitable[Any]]:
    if iscoroutinefunction(func) and callable(func):
        return func

    if callable(func):

        async def async_func(*args, **kwargs):
            return func(*args, **kwargs)

        return async_func

    if isawaitable(func):

        async def async_await_func(*_args, **_kwargs):
            return await func

        return async_await_func

    async def async_return_func(*_args, **_kwargs):
        return func

    return async_return_func


class AsyncWrapperMixin:
    async def wrap_async_exit_stack(
        self, value: Any, wrap_config: Union[bool, List[str]]
    ) -> Tuple[Optional[AsyncExitStack], Any]:
        exit_stack = AsyncExitStack()
        try:
            if isinstance(wrap_config, bool):
                if not wrap_config:
                    return None, value
                wrapped_value = await exit_stack.enter_async_context(value)
                return exit_stack, wrapped_value

            sentinel = object()

            for attr_path in wrap_config:
                attr_path = attr_path.strip()
                check_missing_values = True
                if attr_path.endswith(":ignore_missing"):
                    check_missing_values = False
                    attr_path.rstrip(":ignore_missing")
                attribute_value = extract_from_obj(
                    obj=value,
                    attribute_path=attr_path,
                    check_attribute_exists=check_missing_values,
                    default_value=sentinel,
                )

                if attribute_value is not sentinel:
                    wrapped_attr_val = await exit_stack.enter_async_context(attribute_value)
                    set_value_obj(obj=value, attribute_path=attr_path, value=wrapped_attr_val)

            return exit_stack, value
        except:
            await exit_stack.aclose()
            raise
